from faster_whisper import WhisperModel
# TODO: What if 'whisper_timestamped' is better? We should use it
import whisper_timestamped


def get_transcription_text(audio_filename, model = 'base'):
    """
    Receives an 'audio_filename' and gets the transcription of it,
    returning only the text. This is a simple and faster method
    to be used when we simply want the text without timestamps.
    """
    model = WhisperModel(model)

    segments, info = model.transcribe(audio_filename, language = 'es', beam_size = 5)

    text = ""
    for segment in segments:
        text += segment.text + ' '

    text = text.strip()

    return text

# Very interesting (https://github.com/m-bain/whisperX) to get timestamps of each word
# But the one I use is this one (https://github.com/linto-ai/whisper-timestamped)
def get_transcription_with_timestamps(audio_filename, model = 'small', initial_prompt = None):
    """
    Receives an 'audio_filename' and makes a transcription of it, returning a 
    list of 'segments' in the result. You can pass an 'initial_prompt' to help
    the model to recognize some specific words and to have a better performance.

    This method returns a list of the words recognized in the audio with the
    next parameters: 'text' (word string), 'start' (the moment in which that
    word starts being dictated in the audio), 'end' (the moment in which the 
    word stops being dictated in the audio) and 'confidende' (the certainty
    with which the model has recognize that word, from 0 to 1).
    """
    audio = whisper_timestamped.load_audio(audio_filename)
    model = whisper_timestamped.load_model(model, device = "cpu")
    
    # I do this 'initial_prompt' formatting due to some issues when using it 
    # as it was. I read in Cookguide to pass natural sentences like this below
    # and it seems to be working well, so here it is:
    if initial_prompt != None:
        #initial_prompt = '""""' + initial_prompt + '""""'
        #initial_prompt = 'I know exactly what is said in this audio and I will give it to you (between double quotes) to give me the exact transcription. The audio says, exactly """"' + initial_prompt + '""""'
        initial_prompt = 'I will give you exactly what the audio says (the output), so please ensure it fits. The output must be """"' + initial_prompt + '""""'

    # 'vad' = True would remove silent parts to decrease hallucinations
    # 'detect_disfluences' detects corrections, repetitions, etc. so the
    # word prediction should be more accurate. Useful for natural narrations
    result = whisper_timestamped.transcribe(model, audio, language = "es", initial_prompt = initial_prompt)
    
    return result


# TODO: Remove this when new greenscreen system has been
# completely implemented and this file doesn't make sense
from yta_general_utils.temp import create_temp_filename
from yta_general_utils.color.converter import rgb_to_hex
from PIL import Image, ImageDraw, ImageFont
from random import choice as random_choice
from moviepy.editor import VideoFileClip, ImageClip, CompositeVideoClip, vfx

import os


# TODO: This has to be refactored to be dynamically built
# and to obtain the files from Google Drive, that must be
# also registered in somewhere to include details like the
# size, the number of green screens and their positions, 
# etc.
#
# We will need to refactor it again to handle different
# texts (not only one) and more elements that could be
# applied while whe develop the app an this greenscreen
# functionality.
class TitledGreenScreenImage:
    """
    Green screen of custom size that is placed over a background image of 1920x1080. It
    can include a text that is placed in the bottom side of the green screen.
    """
    __TOUSE_ABSOLUTE_PATH = os.getenv('TOUSE_ABSOLUTE_PATH')
    __GREEN_SCREENS_FOLDER = __TOUSE_ABSOLUTE_PATH + 'green_screens/'
    # TODO: This is no more a local folder, they are files that we must
    # download from Google Drive
    __BACKGROUNDS_1920x1080_FOLDER = __GREEN_SCREENS_FOLDER + 'backgrounds/1920x1080/'
    __GREEN_SCREEN_RGB_COLOR = (0, 249, 12)
    # TODO: Move this to .env
    __FONTS_PATH = 'C:/USERS/DANIA/APPDATA/LOCAL/MICROSOFT/WINDOWS/FONTS/'

    def __init__(self, title, width = 1344, height = 756):
        if width <= 0 or width > 1920 or height <= 0 or height > 1080:
            # TODO: Handle this exception
            print('Green screen size error, setting to default (1344x756).')
            width = 1344
            height = 756

        # TODO: We must now download the file and use it
        self.background_filename = self.__BACKGROUNDS_1920x1080_FOLDER + random_choice(os.listdir(self.__BACKGROUNDS_1920x1080_FOLDER))
        self.title = title
        self.width = width
        self.height = height

        self.x = int((Image.open(self.background_filename).width - self.width) / 2)
        self.y = int((Image.open(self.background_filename).height - self.height) / 2)

    def save(self, output_filename):
        # We use a random 1920x1080 background image
        # We insert the green_screen in the base_image
        base = Image.open(self.background_filename)
        green_screen = Image.open(self.__generate_green_screen())

        # Dynamically get x,y start to put the green screen in the base image center
        base.paste(green_screen, (self.x, self.y))

        PADDING = 15

        if self.title:
            draw = ImageDraw.Draw(base)   
            font = ImageFont.truetype(self.__FONTS_PATH + 'ROBOTO-MEDIUM.TTF', 60, encoding = 'unic')
            w, h = draw.textsize(self.title, font = font)
            position = ((1920 / 2) - (w / 2), 1080 - ((1080 - self.height) / 2) - (h / 2 + PADDING))
            left, top, right, bottom = draw.textbbox(position, self.title, font = font)
            draw.rectangle((left - PADDING * 2, top - PADDING, right + PADDING * 2, bottom + PADDING), fill = 'black')
            draw.text(position, self.title, font = font, fill = 'white')
        
        base.save(output_filename, quality = 100)

    def __generate_green_screen(self):
        """
        Generates a green screen rectangle of the provided 'width' and 'height' that has a rectangle
        black border.
        """
        shape = [(10, 10), (self.width - 10, self.height - 10)] 
        
        img = Image.new("RGB", (self.width, self.height)) 
        draw = ImageDraw.Draw(img)   
        r, g, b = self.__GREEN_SCREEN_RGB_COLOR
        draw.rectangle(shape, fill = rgb_to_hex(r, g, b) , outline = 'black')

        filename = create_temp_filename('tmp_gs_shape.png')
        img.save(filename)

        return filename
    
    def insert_video(self, video_filename, output_filename):
        tmp_filename = create_temp_filename('tmp_gs.png')
        self.save(tmp_filename)

        clip = VideoFileClip(video_filename)
        green_screen_clip = ImageClip(tmp_filename, duration = clip.duration).fx(vfx.mask_color, color = self.rgb_color, thr = 100, s = 5)

        width = self.drx - self.ulx
        clip = clip.resize(width = width).set_position((self.ulx, self.uly))

        final_clip = CompositeVideoClip([clip, green_screen_clip], size = green_screen_clip.size)

        final_clip.write_videofile(output_filename, fps = clip.fps)

    def insert_image(self, image_filename, output_filename):
        # I do the trick with moviepy that is working for videos...
        tmp_filename = create_temp_filename('tmp_gs.png')
        self.save(tmp_filename)

        width = self.drx - self.ulx

        clip = ImageClip(image_filename, duration = 1 / 60).resize(width = width).set_position((self.ulx, self.uly))
        green_screen_clip = ImageClip(tmp_filename, duration = clip.duration).fx(vfx.mask_color, color = self.rgb_color, thr = 100, s = 5)

        
        final_clip = CompositeVideoClip([clip, green_screen_clip], size = green_screen_clip.size)
        final_clip.save_frame(output_filename, t = 0)

        # https://medium.com/@gowtham180502/how-can-we-replace-the-green-screen-background-using-python-4947f1575b1f
        """
        # https://www.geeksforgeeks.org/replace-green-screen-using-opencv-python/
        # TODO: Please, implement with this below, but by now (above)
        image = cv2.imread(self.__filename) 
        frame = cv2.imread(output_filename)
    
        width = self.drx - self.ulx
        height = self.dry - self.uly
        frame = resize_image(resize_image, width, height)

        u_green = np.array([104, 153, 70]) 
        l_green = np.array([30, 30, 0]) 
    
        mask = cv2.inRange(frame, l_green, u_green) 
        res = cv2.bitwise_and(frame, frame, mask = mask) 
    
        f = frame - res 
        f = np.where(f == 0, image, f) 
    
        cv2.imshow("video", frame) 
        cv2.imshow("mask", f) 
        """
    



from moviepy.editor import AudioFileClip

import sox


def concatenate_audio_files(audio_filenames: list[str], output_filename: str):
    """
    Receives a set of audio filenames and turns them into a single file. This
    method writes the new file as 'output_filename' and returns the new file
    as an AudioFileClip.
    """
    if not output_filename:
        return None
    
    if not audio_filenames:
        return None
    # TODO: Check if all 'audio_filenames' provided are valid

    cbn = sox.Combiner()
    cbn.set_input_format(["mp3"] * len(audio_filenames), rate = [44100] * len(audio_filenames), channels = [2] * len(audio_filenames))
    cbn.convert(samplerate = 44100, n_channels = 2, bitdepth = 64)
    cbn.build(audio_filenames, output_filename, "concatenate")

    return AudioFileClip(output_filename)

def concatenate_audioclips(clips: list[AudioFileClip], output_filename: str):
    """
    Receives a set of AudioFileClips and turns them into a single file. This
    method writes the new file as 'output_filename' and returns the new file
    as an AudioFileClip.
    """
    if not output_filename:
        return None
    
    if not clips:
        return None
    # TODO: Check if all 'audio_filenames' provided are valid

    cbn = sox.Combiner()
    cbn.set_input_format(["mp3"] * len(clips), rate = [44100] * len(clips), channels = [2] * len(clips))
    cbn.convert(samplerate = 44100, n_channels = 2, bitdepth = 64)
    filenames = [clip.filename for clip in clips]
    cbn.build(filenames, output_filename, "concatenate")

    return AudioFileClip(output_filename)


from yta_general_utils.file.reader import read_json_from_file
from yta_general_utils.file.writer import write_json_to_file
from yta_general_utils.path import create_file_abspath
from yta_general_utils.file.checker import file_exists, file_is_video_file
from yta_general_utils.programming.path import get_project_abspath
from yta_general_utils.checker.type import variable_is_type
from yta_general_utils.downloader.google_drive import get_id_from_google_drive_url
from yta_multimedia.resources import get_resource
from yta_multimedia.video.frames import get_frame_from_video_by_frame_number
from moviepy.editor import VideoFileClip, ImageClip, CompositeVideoClip, concatenate_videoclips, vfx
from PIL import Image
from typing import Union, Any

import numpy as np


# TODO: Review all this code please, a lot is duplicated
# TODO: Make this below expect a type 'Image', but 'Image' or
# 'Image.Image' returns each arg must be a type. Got <module 
# 'PIL.Image'. We also can expect np.ndarray when read from
# memory (video.get_frame(...))
def get_greenscreen_area_details(image: Union[str, Any]):
    """
    This methods detects the greenscreen area in the provided
    'image_filename' image that contains one and valid 
    greenscreen.

    This method returns a dictionary containing 'rgb_color', 
    that is the green rgb color of the greenscreen area, and
    'upper_left_pixel' and 'lower_right_pixel', that are the
    greenscreen area corners we need to know. Each of those
    are (x, y) tuples containing the pixel coordinates. The
    'rgb_color' will be a (r, g, b) tuple containing the 
    values from 0 to 255 for each color.

    This method will return None in 'rgb_color' if no green
    color detected. This method will fail if more than one
    separated greenscreens in the provided 'image_filename'.
    """
    # TODO: This is failing when numpy parameter because
    # The truth value of an array with more than one 
    # element is ambiguous. Use a.any() or a.all()
    # if image == None:
    #     return None

    if variable_is_type(image, str):
        # TODO: Check if image is actually an image
        image = Image.open(image)

    if variable_is_type(image, np.ndarray):
        image = numpy_to_pil(image)

    image = image.convert('RGB')
    
    green_rgb_color = get_most_common_rgb_color(image)

    upper_left = {
        'x': 99999,
        'y': 99999,
    }
    lower_right = {
        'x': -1,
        'y': -1,
    }

    for x in range(image.width):
        for y in range(image.height):
            rgb_color = (r, g, b) = image.getpixel((x, y))

            if rgb_color == green_rgb_color:
                if x < upper_left['x']:
                    upper_left['x'] = x
                if y < upper_left['y']:
                    upper_left['y'] = y
                
                if x > lower_right['x']:
                    lower_right['x'] = x
                if y > lower_right['y']:
                    lower_right['y'] = y

    # We apply some margin to make sure we fit the green screen
    MARGIN = 2

    if (upper_left['x'] - MARGIN) > 0:
        upper_left['x'] -= MARGIN
    else:
        upper_left['x'] = 0

    if (upper_left['y'] - MARGIN) > 0:
        upper_left['y'] -= MARGIN
    else:
        upper_left['y'] = 0

    if (lower_right['x'] + MARGIN) < 1920:
        lower_right['x'] += MARGIN
    else:
        lower_right['x'] = 1920

    if (lower_right['y'] + MARGIN) < 1080:
        lower_right['y'] += MARGIN
    else:
        lower_right['y'] = 1080

    return {
        'rgb_color': green_rgb_color,
        'upper_left_pixel': (upper_left['x'], upper_left['y']),
        'lower_right_pixel': (lower_right['x'], lower_right['y'])
    }

# TODO: Move this to a 'yta_general_utils' file
def numpy_to_pil(numpy_image: np.ndarray):
    """
    Turns the 'numpy_image' ndarray to PIL readable image.
    """
    return Image.fromarray((numpy_image * 255).astype(np.uint8))

# TODO: How to set PIL type (?)
# TODO: Move this to a 'yta_general_utils' file
def pil_to_numpy(pil_image):
    """
    Turns the 'pil_image' to a numpy array. The PIL image must
    be an array produced by the code 'Image.open(image_filename)'.
    """
    return np.asarray(pil_image)


# TODO: Make this below expect a type 'Image', but 'Image' or
# 'Image.Image' returns each arg must be a type. Got <module 
# 'PIL.Image'
def get_most_common_rgb_color(image: Union[str, Any], force_green = True):
    """
    Returns the most common green rgb color that exist in the provided
    'image_filename'. There could be no green color so it will return
    None, or the green color as (r, g, b) if existing.

    # TODO: This method could be made more generic by using a range
    # and then specializing this one by using that with green range
    """
    if not image:
        raise Exception('No valid "image" provided')
    
    if variable_is_type(image, str):
        # TODO: Check if image is actually an image
        image = Image.open(image)

    colors = {}
    image = image.convert('RGB')

    # TODO: Maybe improve this code below with this:
    # https://www.projectpro.io/recipes/find-most-frequent-value-array
    #
    # This finds the most common value in numpy array
    # counts = np.bincount(a)
    # most_frequent_value = np.argmax(counts)

    # We will check the most common rgb color (should be the green of mask)
    for x in range(image.width):
        for y in range(image.height):
            rgb_color = (r, g, b) = image.getpixel((x, y))

            if not rgb_color in colors:
                colors[rgb_color] = 1
            else:
                colors[rgb_color] += 1

    # Check which one is the most common
    most_used_rgb_color = {
        'color': None,
        'times': 0,
    }
    for key, value in colors.items():
        if force_green:
            # We only care about green colors
            r, g, b = key
            is_green = (r >= 0 and r <= 100) and (g >= 100 and g <= 255) and (b >= 0 and b <= 100)
            if is_green:
                if value > most_used_rgb_color['times']:
                    most_used_rgb_color = {
                        'color': key,
                        'times': value
                    }
        else:
            if value > most_used_rgb_color['times']:
                most_used_rgb_color = {
                    'color': key,
                    'times': value
                }

    return most_used_rgb_color['color']



"""
        Moving Greenscreens below
        TODO: Move this below to another more specific file, thank you
        TODO: I'm placing this below here by now, but maybe we should
        build a 'MovingVideoGreenscreen' object for this
"""

# TODO: Move this to a consts.py or something
GREENSCREENS_FOLDER = 'resources/greenscreens/'

def insert_video_into_moving_greescreen(greenscreen_google_drive_url: str, video: Union[VideoFileClip, CompositeVideoClip, ImageClip, str], output_filename: Union[str, None] = None):
    """
    This method inserts the provided 'video' into the moving greenscreen
    that is contained in the provided 'greenscreen_google_drive_url'. It
    will extract all greenscreen frames, analyze each of them and extract
    the greenscreen areas information and put the provided 'video' 
    combined with the greenscreen.

    This method will identify the Google Drive ID to store (or load if
    previously downloaded) the moving greenscreen video to be able to
    handle it. It will use one specific folder to work with the data:
    'resources/greenscreens/GOOGLE_DRIVE_ID', extracting
    the frames, generating a 'config.json' file and also downloading
    the resource from Google Drive.

    @param
        **greenscreen_google_drive_url**
        The Google Drive url of the moving greenscreen video resource.

    @param
        **video**
        The filename of the video we want to put into the greenscreen.

    @param
        **output_filename**
        The filename in which we want to write the final video that is
        our 'video' inside of the greenscreen.
    """
    if not video:
        raise Exception('No valid "video" provided.')
    
    if variable_is_type(video, str):
        if not file_is_video_file(video):
            raise Exception('Parameter "video" provided is not a valid video file.')
        
        video = VideoFileClip(video)
    
    __preprocess_moving_greenscreen(greenscreen_google_drive_url)

    # Here we have the moving greenscreen with the source file and
    # the config file ready to use
    google_drive_id = get_id_from_google_drive_url(greenscreen_google_drive_url)
    moving_gs_folder = GREENSCREENS_FOLDER + google_drive_id
    greenscreen_filename = moving_gs_folder + '/greenscreen.mp4'
    greenscreen_config_filename = moving_gs_folder + '/config.json'

    moving_gs = VideoFileClip(greenscreen_filename)

    if moving_gs.duration < video.duration:
        video = video.set_duration(moving_gs.duration)
    elif video.duration < moving_gs.duration:
        moving_gs = moving_gs.set_duration(video.duration)

    # We dynamically load main video frames
    frames = []
    for i in range((int) (video.duration * video.fps)):
        frames.append(get_frame_from_video_by_frame_number(video, i))

    greenscreen_details = read_json_from_file(greenscreen_config_filename)

    # Now lets put together greenscreen and main video frames
    # applying a mask to build the final video by concatenating
    # them
    final_clips = []
    for index, frame in enumerate(frames):
        frame_details = greenscreen_details[str(index)]

        # We need to put the image into the greenscreen frame
        imageclip = ImageClip(frame, duration = 1 / 60)
        
        # TODO: Pay atenttion to this:
        # I have a 'to_video' method in ImageGreenscreen that
        # should be used here, but I can't. Why? Because of its 
        # behaviour right now. ImageGreenscreen object is not
        # prepared to have numpy images stored to be used later,
        # so when I call the 'to_video' method I should be able
        # to use the frame I get here below, but I can't because
        # it is expecting an image on Google Drive or a local
        # storaged filename. I don't want to write the frame
        # because of performance decreases, so I leave this text
        # here for the future, to make own custom Greenscreen
        # objects able to handle images and videos loaded in
        # memory to work with.

        # Inject the imageclip
        # TODO: This is a sensitive part, documentation says something
        # about masks when numpy provided, but this is working
        green_screen_clip = ImageClip(get_frame_from_video_by_frame_number(moving_gs, index), duration = imageclip.duration).fx(vfx.mask_color, color = frame_details['rgb_color'], thr = 100, s = 5)
        
        # We cannot resize with numpys
        imageclip = imageclip.resize(width = frame_details['lower_right_pixel'][0] - frame_details['upper_left_pixel'][0]).set_position((frame_details['upper_left_pixel'][0], frame_details['upper_left_pixel'][1]))

        final_clips.append(CompositeVideoClip([imageclip, green_screen_clip], size = green_screen_clip.size))

    final = concatenate_videoclips(final_clips).set_audio(video.audio)

    if output_filename:
        final.write_videofile(output_filename, fps = video.fps)

    return final

def __prepare_moving_greenscreen(google_drive_url: str):
    """
    This method will download (if needed) the resource from Google
    Drive to local and will create (if it doesn't exist) the config
    file of a moving Greenscreen.

    TODO: What about other greenscreens (?)

    @param
        **google_drive_url**
        The Google Drive url of the greenscreen. This will be used
        to download the source file and also to identify it with
        the Google Drive ID.
    """
    if not google_drive_url:
        raise Exception('No "google_drive_url" provided.')

    # The identifier of the greenscreen is the Google Drive id
    google_drive_id = get_id_from_google_drive_url(google_drive_url)
    greenscreen_filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4'
    config_filename = GREENSCREENS_FOLDER + google_drive_id + '/config.json'

    # Check if folder exist or not, and create it if not
    create_file_abspath(get_project_abspath() + greenscreen_filename)

    # TODO: What about image Greenscreen, not .mp4 (?)

    # Check if video exist
    if not file_exists(greenscreen_filename):
        get_resource(google_drive_url, GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4')

    # Check if config file exist
    if not file_exists(config_filename):
        write_json_to_file({}, config_filename)

def __preprocess_moving_greenscreen(greenscreen_google_drive_url: str):
    """
    This method analyzes, frame by frame, a video greenscreen and stores
    each frame information about greenscreen area in a config file to
    be lately used for video injection.
    """
    if not greenscreen_google_drive_url:
        raise Exception('No "greenscreen_google_drive_url" provided.')
    
    # First, make sure config file and source file are available
    __prepare_moving_greenscreen(greenscreen_google_drive_url)

    google_drive_id = get_id_from_google_drive_url(greenscreen_google_drive_url)
    filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4'

    # Read the video and extract frames
    gs_clip = VideoFileClip(filename)

    for i in range((int) (gs_clip.fps * gs_clip.duration)):
        __process_moving_greenscreen_frame(greenscreen_google_drive_url, i)

    return

def __process_moving_greenscreen_frame(greenscreen_google_drive_url: str, frame_number: int, do_force: bool = False):
    """
    This method processes the 'frame_number' frame of the existing
    greenscreen in 'greenscreen_google_drive_url' source. It will
    detect the greenscreen area of that frame and store it in the
    specific configuration file.

    This method will store the information in an specific folder
    'resources/greenscreens/frames/GOOGLE_DRIVE_ID/'. All
    the information is the greenscreen video, the frames and the
    'config.json' file that contains the detailed greenscreen
    areas of each frame. As you can see, the 'GOOGLE_DRIVE_ID' is
    used as identifier.

    If that 'frame_number' has been previously analyzed, the 
    system will do it again only if the 'do_force' parameter
    is True. If not, it will stop and return.

    @param
        **greenscreen_google_drive_url**
        The Google Drive url that contains the greenscreen video
        from which we will analyze the frame.
    
    @param
        **frame_number**
        The frame number that will be analyzed. The first frame of
        the greenscreen is the number 0.

    @param
        **do_force**
        The information of that 'frame_number' could be stored 
        previously so it would be ignored. You can set this
        parameter to True to force recalculating the frame.
    """
    if not greenscreen_google_drive_url:
        return None
    
    if frame_number < 0:
        return None
    
    # We always make sure greenscreen source file and config are ready
    __prepare_moving_greenscreen(greenscreen_google_drive_url)

    # sample url: https://drive.google.com/file/d/1My5V8gKcXDQGGKB9ARAsuP9MIJm0DQOK/view?usp=sharing
    google_drive_id = get_id_from_google_drive_url(greenscreen_google_drive_url)
    
    config_filename = GREENSCREENS_FOLDER + google_drive_id + '/config.json'
    greenscreen_filename = GREENSCREENS_FOLDER + google_drive_id + '/greenscreen.mp4'

    json_data = read_json_from_file(config_filename)

    if str(frame_number) in json_data and not do_force:
        # Frame information found and no forcing
        return

    # We make sure we have the clip (and resource)
    gs_clip = VideoFileClip(greenscreen_filename)

    greenscreen_frames_number = (int) (gs_clip.fps * gs_clip.duration)
    if frame_number > (greenscreen_frames_number):
        raise Exception('The requested "frame_number" is not valid. The greenscreen only has "' + str(greenscreen_frames_number) + '" frames.')
    
    # This below stores the frame and use it later for analysis (working)
    # output_filename = create_temp_filename('tmp_frame.png')
    # get_frame_from_video_by_frame_number(gs_clip, frame_number, output_filename)
    # details = get_greenscreen_area_details(output_filename)

    # This below tries to do with numpy
    # TODO: Maybe this is making it slow (too much memory)
    details = get_greenscreen_area_details(get_frame_from_video_by_frame_number(gs_clip, frame_number))

    json_data[frame_number] = {
        'index': frame_number,
        'rgb_color': details['rgb_color'],
        'upper_left_pixel': details['upper_left_pixel'],
        'lower_right_pixel': details['lower_right_pixel'],
        'frames': None
    }

    return write_json_to_file(json_data, config_filename)


# TODO: By now I'm creating a new category apart from 'video', 
# 'audio' and 'image' because I don't know where can I put this
# code as green screens could be videos or images and they can
# also have images or videos inside


from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from yta_multimedia.greenscreen.custom.utils import get_greenscreen_details
from yta_multimedia.greenscreen.enum.greenscreen_type import GreenscreenType
from yta_general_utils.checker.type import variable_is_type
from yta_general_utils.file.checker import file_is_video_file
from moviepy.editor import VideoFileClip, ImageClip, CompositeVideoClip, vfx, concatenate_videoclips
from typing import Union

class VideoGreenscreen:
    greenscreen: GreenscreenDetails = None
    greenscreen_clip = None
    """
    The last greenscreen clip used to build a video. This clip will
    have the actual used greenscreen resource duration, so you can
    now how long is the greenscreen actually. You can use this clip
    to obtain the 'duration' when the 'from_image_to_video' or 
    'from_video_to_video' method is used so you can now how much 
    time is actually on greenscreen and how much is left because of
    the resource duration.

    You must remember that if a greenscreen of 3 seconds is applied
    over a video of 5 seconds of duration, only the 3 first secconds
    will be with the greenscreen, and the other 2 seconds will remain
    as in the original video.
    """

    def __init__(self, greenscreen: Union[GreenscreenDetails, str]):
        if variable_is_type(greenscreen, str):
            # We need to automatically detect greenscreen details
            greenscreen = get_greenscreen_details(greenscreen, GreenscreenType.VIDEO)

        self.greenscreen = greenscreen
       
    def from_image_to_image(self):
        # TODO: Does this make sense? We are working with a video, so
        # trying to generate just a frame, an image, does make sense?
        # TODO: By the way, this is not working yet
        pass

    def from_image_to_video(self, image_filename: str, duration: float = 3.0, output_filename: Union[str, None] = None):
        """
        Receives an 'image_filename', places it into the greenscreen and generates
        a video of 'duration' seconds of duration that is returned. This method
        will store locally the video if 'output_filename' is provided.
        """
        if not image_filename:
            return None
        
        if not duration:
            return None
        
        if not output_filename:
            return None
        
        imageclip = ImageClip(image_filename, duration = duration)
        final_clip = self.from_video_to_video(imageclip)

        if output_filename:
            final_clip.write_videofile(output_filename)

        return final_clip
    
    def from_video_to_image(self):
        # TODO: Does this make sense? By now it is not implemented
        pass

    def from_video_to_video(self, video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], output_filename: Union[str, None] = None):
        """
        Inserts the provided 'video' in the greenscreen and returns the
        CompositeVideoClip that has been created. If 'output_filename' 
        provided, it will be written locally with that file name.

        The provided 'video' can be a filename or a moviepy video clip.

        TODO: Careful when video is longer than greenscreen
        """
        if not video:
            return None
        
        if variable_is_type(video, str):
            if not file_is_video_file(video):
                return None
            
            video = VideoFileClip(video)

        # TODO: Change this 
        # We could have different areas in which append 1 or more 
        # videos, so it would be handled different. By now I'm
        # considering we only have one

        TMP_FILENAME = self.greenscreen.get_filename()

        # We choose the first greenscreen area to work with
        greenscreen_area = self.greenscreen.greenscreen_areas[0]
        
        # Create the clip
        self.green_screen_clip = VideoFileClip(TMP_FILENAME).fx(vfx.mask_color, color = greenscreen_area.rgb_color, thr = 100, s = 5)

        width = greenscreen_area.lower_right_pixel[0] - greenscreen_area.upper_left_pixel[0]
        # If the provided clip is shorter than our green screen: easy, crop the green screen
        # If the provided clip is longer than our green screen: I use the green screen duration
        # and let the clip be original the rest of the time
        # TODO: Improve resizing process as it is failing...
        if self.green_screen_clip.duration > video.duration:
            self.green_screen_clip = self.green_screen_clip.set_duration(video.duration)
            # Clip will be displayed inside the green screen area
            video = video.resize(width = width).set_position((greenscreen_area.upper_left_pixel[0], greenscreen_area.upper_left_pixel[1]))
            video = CompositeVideoClip([video, self.green_screen_clip], size = self.green_screen_clip.size)
        elif video.duration > self.green_screen_clip.duration:
            # First subclip will be displayed inside the green screen area
            first_clip = video.subclip(0, self.green_screen_clip.duration).resize(width = width).set_position((greenscreen_area.upper_left_pixel[0], greenscreen_area.upper_left_pixel[1]))
            # Second clip will be as the original one
            second_clip = video.subclip(self.green_screen_clip.duration, video.duration)
            video = concatenate_videoclips([
                CompositeVideoClip([first_clip, self.green_screen_clip], size = self.green_screen_clip.size),
                second_clip
            ])
        else:
            video = CompositeVideoClip([video, self.green_screen_clip], size = self.green_screen_clip.size)

        if output_filename:
            video.write_videofile(output_filename, fps = video.fps)

        return video


from yta_multimedia.greenscreen.classes.greenscreen_details import GreenscreenDetails
from yta_multimedia.greenscreen.custom.utils import get_greenscreen_details
from yta_multimedia.video.utils import parse_parameter_as_moviepy_clip
from yta_multimedia.greenscreen.enum.greenscreen_type import GreenscreenType
from yta_general_utils.checker.type import variable_is_type
from moviepy.editor import ImageClip
from typing import Union
from moviepy.editor import VideoFileClip, ImageClip, CompositeVideoClip, vfx
from PIL import Image, ImageDraw
from typing import Union, Any

import numpy as np


class ImageGreenscreen:
    greenscreen: GreenscreenDetails = None

    def __init__(self, greenscreen: Union[GreenscreenDetails, str]):
        if variable_is_type(greenscreen, str):
            # We need to automatically detect greenscreen details
            greenscreen = get_greenscreen_details(greenscreen, GreenscreenType.IMAGE)

        self.greenscreen = greenscreen

    def __process_elements_and_save(self, output_filename):
        """
        Processes the greenscreen by writing the title, description
        and any other available element, and stores it locally as
        'output_filename' once processed.
        """
        base = Image.open(self.greenscreen.filename_or_google_drive_url)
        draw = ImageDraw.Draw(base)

        # TODO: I preserve this code for the future
        # # We need to write title if existing
        # if self.__title:
        #     title_position = (self.__title_x, self.__title_y)
        #     draw.text(title_position, self.__title, font = self.__title_font, fill = self.__title_color)

        # if self.__description:
        #     description_position = (self.__description_x, self.__description_y)
        #     draw.text(description_position, self.__description, font = self.__description_font, fill = self.__description_color)

        # TODO: Handle anything else here

        # We save the image
        base.save(output_filename, quality = 100)

    # TODO: Make this below expect a type 'Image', but 'Image' or
    # 'Image.Image' returns each arg must be a type. Got <module 
    # 'PIL.Image'
    def from_image_to_image(self, image: Union[str, Any], output_filename: Union[str, None] = None):
        """
        Receives an 'image_filename', places it into the greenscreen and generates
        another image that is stored locally as 'output_filename'.
        """
        # TODO: This is failing when numpy parameter because
        # The truth value of an array with more than one 
        # element is ambiguous. Use a.any() or a.all()
        # if image == None:
        #     return None
        
        if not output_filename:
            return None
        
        if variable_is_type(image, str):
            # TODO: Check if image is valid

            image = np.asarray(Image.open(image))
        
        # Put the image into the greenscreen
        imageclip = ImageClip(image, duration = 1 / 60)
        final_clip = self.from_video_to_video(imageclip)

        # Generate the result image
        if output_filename:
            final_clip.save_frame(output_filename, t = 0)

        return final_clip.get_frame(t = 0)

    def from_image_to_video(self, image: str, duration: float = 3.0, output_filename: Union[str, None] = None):
        """
        Receives an 'image', places it into the greenscreen and generates
        a video of 'duration' seconds of duration that is returned. This method
        will store locally the video if 'output_filename' is provided.
        """
        if not image:
            return None
        
        if not duration:
            return None
        
        if not output_filename:
            return None
        
        if variable_is_type(image, str):
            # TODO: Check if image is valid

            image = np.asarray(Image.open(image))

        imageclip = ImageClip(image, duration = duration)
        final_clip = self.from_video_to_video(imageclip)

        if output_filename:
            final_clip.write_videofile(output_filename)

        return final_clip
    
    def from_video_to_image(self):
        # TODO: Does this make sense? By now it is not implemented
        pass

    def from_video_to_video(self, video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip], output_filename: Union[str, None] = None):
        """
        Inserts the provided 'video' in the greenscreen and returns the
        CompositeVideoClip that has been created. If 'output_filename' 
        provided, it will be written locally with that file name.

        The provided 'video' can be a filename or a moviepy video clip.

        TODO: Careful when video is longer than greenscreen
        """
        video = parse_parameter_as_moviepy_clip(video)

        # TODO: Change this 
        # We could have different areas in which append 1 or more 
        # videos, so it would be handled different. By now I'm
        # considering we only have one
        
        # We download the file to be able to use it
        # TODO: This should be refactor with new fields in ImageGreenscreen
        # such as 'filename' and 'google_drive_url' by separate
        # TODO: We should check if this is a filename or a Google Drive url
        # and act in consecuence
        #TMP_FILENAME = get_resource(self.greenscreen.filename_or_google_drive_url, self.greenscreen.filename_or_google_drive_url)

        # If it is a filename we cannot get the id
        TMP_FILENAME = self.greenscreen.get_filename()

        # We process greenscreen elements if needed
        # TODO: This will be done in the future (writing titles)
        #tmp_filename = create_temp_filename('tmp_gs.png')
        #self.__process_elements_and_save(tmp_filename)

        # We choose the first greenscreen area to work with
        greenscreen_area = self.greenscreen.greenscreen_areas[0]

        # We work to maintain the proportion
        ulx = greenscreen_area.upper_left_pixel[0]
        uly = greenscreen_area.upper_left_pixel[1]
        lrx = greenscreen_area.lower_right_pixel[0]
        lry = greenscreen_area.lower_right_pixel[1]
        
        # TODO: Please, review this below
        gs_width = lrx - ulx
        gs_height = lry - uly
        aspect_ratio = gs_width / gs_height
        if aspect_ratio > (1920 / 1080):
            # We keep the width, but we calculate the height to keep 16/9 aspect ratio
            gs_height_for_16_9 = gs_width * (1080 / 1920)
            # Now we know the new height, so we only have to center the video on 'y' axis
            difference = gs_height_for_16_9 - gs_height
            # My video (16/9) will not fit the whole width
            uly -= difference / 2
            lry += difference / 2
        elif aspect_ratio < (1920 / 1080):
            # We keep the height, but we calculate the width to keep 16/9 aspect ratio
            gs_width_for_16_9 = gs_height * (1920 / 1080)
            # Now we know the new width, so we only have to center the video on 'x' axis
            difference = gs_width_for_16_9 - gs_width
            ulx -= difference / 2
            lrx += difference / 2
        # TODO: Are we sure that those coords are int and make a real 16/9 with int numbers?
        
        # Create the clip
        green_screen_clip = ImageClip(TMP_FILENAME, duration = video.duration).fx(vfx.mask_color, color = greenscreen_area.rgb_color, thr = 100, s = 5)
        width = lrx - ulx
        video = video.resize(width = width).set_position((ulx, uly))

        final_clip = CompositeVideoClip([video, green_screen_clip], size = green_screen_clip.size)

        if output_filename:
            final_clip.write_videofile(output_filename, fps = video.fps)

        return final_clip


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from yta_multimedia.video.frames import get_all_frames_from_video
from yta_general_utils.temp import create_temp_filename
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, AudioFileClip, ImageSequenceClip, ColorClip, VideoClip
from typing import Union
from pydub import AudioSegment


class ReversedVideoEffect(VideoEffect):
    """
    This method creates a new one but in reversa, also with the sound reversed.

    It doesn't use the 'mirror_time' effect because it fails. Instead, it saves
    each frame of the video and builds a new video using them in reverse order.
    It also uses the original audio an reverses it in the new generated video.
    """

    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        Applies the effect to the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        reversed_frames_array = get_all_frames_from_video(video)[::-1]

        # TODO: Try to do this in memory
        AUDIO_FILE = create_temp_filename('tmp_audio.mp3')
        REVERSED_AUDIO_FILE = create_temp_filename('tmp_reversed_audio.mp3')
        video.audio.write_audiofile(AUDIO_FILE, fps = 44100)
        AudioSegment.from_mp3(AUDIO_FILE).reverse().export(REVERSED_AUDIO_FILE)
        reversed_audio = AudioFileClip(REVERSED_AUDIO_FILE)

        return ImageSequenceClip(reversed_frames_array, fps = video.fps).set_audio(reversed_audio)


from yta_general_utils.web.scrapper.chrome_scrapper import ChromeScrapper
from yta_general_utils.temp import create_temp_filename
from yta_multimedia.resources import get_resource
from selenium.webdriver.common.by import By
from PIL import Image
from random import randrange
from typing import Union
from time import strftime, gmtime

class YoutubeImageGenerator:
    def generate_comment(self, author: str, avatar_url: str, time: str, message: str, likes_number: int, output_filename: Union[str, None] = None):
        """
        This method generates a Youtube comment image with the provided
        information. It will return the image read with PIL, but will
        also store the screenshot (as this is necessary while processing)
        with the provided 'output_filename' if provided, or with as a
        temporary file if not.
        """
        if not author:
            # TODO: Fake author name (start with @)
            pass

        if not avatar_url:
            # TODO: Fake avatar_url or just let the one existing
            pass

        if not time:
            # TODO: Fake time ('hace X años,meses,dias,horas')
            pass

        if not message:
            # TODO: Fake a message with AI
            pass

        if not likes_number:
            likes_number = randrange(50)

        scrapper = ChromeScrapper(False)
        # We go to this specific video with comments available
        scrapper.go_to_web_and_wait_util_loaded('https://www.youtube.com/watch?v=OvUj2WsADjI')
        # We need to scroll down to let the comments load
        # TODO: This can be better, think about a more specific strategy
        # about scrolling
        scrapper.scroll_down(1000)
        scrapper.wait(1)
        scrapper.scroll_down(1000)
        scrapper.wait(1)

        # We need to make sure the comments are load
        scrapper.find_element_by_element_type_waiting('ytd-comment-thread-renderer')
        comments = scrapper.find_elements_by_element_type('ytd-comment-thread-renderer')

        comment = comments[3]
        body = comment.find_element(By.ID, 'body')

        # Change user (avatar) image
        image = body.find_element(By.ID, 'img')
        scrapper.set_element_attribute(image, 'src', avatar_url)
        # TODO: Check that Image actually changes in the view
        # maybe with this: https://stackoverflow.com/questions/44286061/how-to-check-that-the-image-was-changed-if-therere-no-changes-in-html-code

        # Change date
        time_element = body.find_element(By.ID, 'published-time-text')
        time_element = scrapper.find_element_by_element_type('a', time_element)
        scrapper.set_element_inner_text(time_element, time)

        # Change user name
        author_element = body.find_element(By.ID, 'header-author')
        author_element = scrapper.find_element_by_element_type('h3', author_element)
        author_element = scrapper.find_element_by_element_type('a', author_element)
        author_element = scrapper.find_element_by_element_type('span', author_element)
        scrapper.set_element_inner_text(author_element, author)

        # Change message
        message_element = scrapper.find_element_by_id('content-text', comment)
        message_element = scrapper.find_element_by_element_type('span', message_element)
        scrapper.set_element_inner_text(message_element, message)

        # Change number of likes
        likes_element = scrapper.find_element_by_id('vote-count-middle', comment)
        scrapper.set_element_inner_text(likes_element, str(likes_number))
        
        scrapper.scroll_to_element(comment)
        
        filename = output_filename
        if not filename:
            filename = create_temp_filename('tmp_comment_screenshot.png')
        
        style = 'width: 500px; padding: 10px;'
        scrapper.set_element_style(comment, style)
        scrapper.screenshot_element(comment, filename)

        return Image.open(filename)

    def generate_thumbnail(self, title: str = None, image_url: str = None, channel_name: str = None, channel_image_url: str = None, duration_in_seconds: int = None, views: str = None, time_since_publication: str = None, output_filename: Union[str, None] = None):
        if not title:
            # TODO: Fake it
            title = 'Título de la miniatura'

        if not image_url:
            # TODO: Fake it
            image_url = 'https://static-cse.canva.com/blob/1697393/1600w-wK95f3XNRaM.jpg'

        if not channel_name:
            # TODO: Fake it
            channel_name = 'Youtube Autónomo'

        if not channel_image_url:
            # TODO: Fake it
            channel_image_url = 'https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcTmIVOqsYK3t8HxkQ_WjwPoP2cwJiV1xDyWIw&s'

        if not duration_in_seconds:
            duration_in_seconds = randrange(300, 700)

        if not views:
            # TODO: Fake it
            views = '2.3M visitas'

        if not time_since_publication:
            # TODO: Fake it
            time_since_publication = 'hace 2 horas'

        scrapper = ChromeScrapper()
        # Go to https://thumbnailchecker.com/es
        scrapper.go_to_web_and_wait_util_loaded('https://thumbnailchecker.com/es')

        # Fill info (we need it to be able to see the modal)
        title_input = scrapper.find_element_by_class('input', 'form_input')
        title_input.send_keys('Título de prueba')
        upload_input = scrapper.find_element_by_id('upload')
        filename = get_resource('https://drive.google.com/file/d/1rcowE61X8c832ynh0xOt60TH1rJfcJ6z/view?usp=drive_link', 'base_thumbnail.png')
        scrapper.set_file_input(upload_input, filename)

        button = scrapper.find_element_by_text_waiting('button', 'Revisa tu miniatura')
        scrapper.scroll_to_element(button)
        button.click()

        thumbnail_container_element = scrapper.find_element_by_class_waiting('div', 'yt_main_box')

        # Image
        image = scrapper.find_element_by_class('div', 'yt_box_thumbnail', thumbnail_container_element)
        image = scrapper.find_element_by_element_type('div', image)
        image = scrapper.find_element_by_element_type('img', image)
        scrapper.set_element_attribute(image, 'src', image_url)
        # TODO: Wait loading (?)

        # Avatar image
        avatar_image = scrapper.find_element_by_class('div', 'yt_box_info_avatar', thumbnail_container_element)
        avatar_image = scrapper.find_element_by_element_type('img', avatar_image)
        scrapper.set_element_attribute(avatar_image, 'src', channel_image_url)
        # TODO: Wait loading (?)

        # Video duration (in 'MM:SS' format)
        video_duration_element = scrapper.find_element_by_class('span', 'yt_time_status')
        duration_str = strftime('%M:%S', gmtime(duration_in_seconds))
        if duration_in_seconds >= 3600:
            duration_str = strftime('%H:%M:%S', gmtime(duration_in_seconds))
            if duration_in_seconds < 36000:
                # We need to set hour as one only digit
                duration_str = duration_str[1:]
        scrapper.set_element_inner_text(video_duration_element, duration_str)

        # Title
        description = scrapper.find_element_by_class('div', 'yt_box_info_content', thumbnail_container_element)
        title_element = scrapper.find_element_by_element_type('h4', description)
        scrapper.set_element_inner_text(title_element, title)

        # Author
        user = scrapper.find_element_by_element_type('p', description)
        scrapper.set_element_inner_text(user, channel_name)

        # Views
        views_element = scrapper.find_element_by_class('div', 'yt_box_info_meta', thumbnail_container_element)
        ul = scrapper.find_element_by_element_type('ul', views_element)
        listed_items = scrapper.find_elements_by_element_type('li', ul)
        # TODO: Handle 'views' and 'time_since_publication' with int
        # and format it manually here
        scrapper.set_element_inner_text(listed_items[1], views)
        scrapper.set_element_inner_text(listed_items[2], time_since_publication)

        style = 'width: 500px; padding: 10px;'
        scrapper.set_element_style(thumbnail_container_element, style)

        if not output_filename:
            output_filename = create_temp_filename('tmp_youtube_thumbnail.png')

        scrapper.screenshot_element(thumbnail_container_element, output_filename)

        return Image.open(output_filename)


from yta_general_utils.checker.type import variable_is_type
from typing import Union

import cv2
import numpy as np
import skimage.exposure


def remove_greenscreen_from_image(image_filename: Union[str, cv2.typing.MatLike], output_filename: str):
    """
    This methods gets a greenscreen image (one with lots of uniform
    green pixels), does some modifications and replaces the green
    pixels with transparent pixels (due to alpha channel).
    """
    # Found here: https://stackoverflow.com/questions/51719472/remove-green-background-screen-from-image-using-opencv-python
    if not image_filename:
        return None
    
    img = image_filename
    if variable_is_type(image_filename, str):
        # TODO: Implement 'file_is_image_file' when available
        img = cv2.imread(image_filename)

    # TODO: This is not working, it raises
    # TypeError: Subscripted generics cannot be used with class and instance checks
    # if not variable_is_type(img, cv2.typing.MatLike):
    #     # TODO: Raise Exception
    #     return None

    # 1. Perform a basic masking
    # Convert the image to LAB space
    lab = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)

    # Threshold the alpha channel to isolate green background
    a_channel = lab[:, :, 1]
    th = cv2.threshold(a_channel, 127, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]

    # Mask the original image with binary mask
    masked = cv2.bitwise_and(img, img, mask = th)    # contains dark background
    m1 = masked.copy()
    m1[th == 0] = (255, 255, 255)                    # contains white background

    # 2. Remove green shade along the border
    # Convert masked image to LAB space
    mlab = cv2.cvtColor(masked, cv2.COLOR_BGR2LAB)

    # Normalize the alph achannel mlab[:, :, 1] to use the 
    # entire intensity range between [0 - 255]
    dst = cv2.normalize(mlab[:, :, 1], dst = None, alpha = 0, beta = 255, norm_type = cv2.NORM_MINMAX, dtype = cv2.CV_8U)

    # Green color represents the lower end of the range [0-255] 
    # while red color represents the higher end in the a-channel
    threshold_value = 100
    dst_th = cv2.threshold(dst, threshold_value, 255, cv2.THRESH_BINARY_INV)[1]

    # Set intensity value in the a-channel of the selected pixels to 127
    mlab2 = mlab.copy()
    mlab[:, :, 1][dst_th == 255] = 127

    # Convert the image to BGRA and the pixels that were dark we 
    # transform to green but with max alpha to be transparent
    img2 = cv2.cvtColor(mlab, cv2.COLOR_LAB2BGR)
    img2 = cv2.cvtColor(img2, cv2.COLOR_BGR2BGRA)
    # We add the alpha channel
    img2[th == 0] = (0, 255, 0, 0)

    if output_filename:
        cv2.imwrite(output_filename, img2)

    # Check this, is interesting: https://stackoverflow.com/questions/44290606/alpha-masking-a-non-square-region-python-cv2

    return img
    
def remove_greenscreen_from_image_with_blur(image_filename: str, output_filename: str):
    """
    This methods gets a greenscreen image (one with lots of uniform
    green pixels), does some modifications and replaces the green
    pixels with transparent pixels (due to alpha channel).

    The result is worse in the test cases than the 
    'remove_greenscreen_from_image' method.
    """
    # Found here: https://stackoverflow.com/questions/51719472/remove-green-background-screen-from-image-using-opencv-python
    if not image_filename:
        return None

    img = image_filename
    if variable_is_type(image_filename, str):
        # TODO: Implement 'file_is_image_file' when available
        img = cv2.imread(image_filename)

    # TODO: This is not working, it raises
    # TypeError: Subscripted generics cannot be used with class and instance checks
    # if not variable_is_type(img, cv2.typing.MatLike):
    #     # TODO: Raise Exception
    #     return None

    # Convert to LAB
    lab = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)

    # Extract A channel
    A = lab[:, :, 1]

    # Threshold A channel
    thresh = cv2.threshold(A, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]

    # Blur threshold image
    blur = cv2.GaussianBlur(thresh, (0, 0), sigmaX = 5, sigmaY = 5, borderType = cv2.BORDER_DEFAULT)

    # Stretch so that 255 -> 255 and 127.5 -> 0
    mask = skimage.exposure.rescale_intensity(blur, in_range = (127.5, 255), out_range = (0, 255)).astype(np.uint8)

    # Add mask to image as alpha channel
    result = img.copy()
    result = cv2.cvtColor(img, cv2.COLOR_BGR2BGRA)
    result[:, :, 3] = mask

    if output_filename:
        cv2.imwrite(output_filename, result)

    return result


from subprocess import run


def remove_image_background(image_filename, output_filename):
    """
    Removes the background of the provided 'image_filename' by using the 
    'backgroundremover' open library that is included in a comment.
    """
    # It uses (https://github.com/nadermx/backgroundremover?referral=top-free-background-removal-tools-apis-and-open-source-models)
    # That uses U2Net (https://medium.com/axinc-ai/u2net-a-machine-learning-model-that-performs-object-cropping-in-a-single-shot-48adfc158483)
    # TODO: Please, use this as python library not as command, ty
    command_parameters = ['backgroundremover', '-i', image_filename, '-o', output_filename]
    run(command_parameters)

    """ # Problem with Circular import
    from backgroundremover.bg import remove as remove_background
    r = lambda image_filename: image_filename.buffer.read() if hasattr(image_filename, "buffer") else image_filename.read()
    w = lambda o, data: o.buffer.write(data) if hasattr(o, "buffer") else o.write(data)

    # These below are default values
    x = remove_background(
        r(image_filename),
        model_name = 'u2net',
        alpha_matting = False,
        alpha_matting_foreground_threshold = 240,
        alpha_matting_background_threshold = 10,
        alpha_matting_erode_structure_size = 10,
        alpha_matting_base_size = 1000
    )
    w(output_filename, x)
    """

    

    # TODO: This below seems to work (as shown in this 
    # commit https://github.com/nadermx/backgroundremover/commit/c590858de4c7e75805af9b8ecdd22baf03a1368f)
    """
    from backgroundremover.bg import remove
    def remove_bg(src_img_path, out_img_path):
        model_choices = ["u2net", "u2net_human_seg", "u2netp"]
        f = open(src_img_path, "rb")
        data = f.read()
        img = remove(data, model_name=model_choices[0],
                    alpha_matting=True,
                    alpha_matting_foreground_threshold=240,
                    alpha_matting_background_threshold=10,
                    alpha_matting_erode_structure_size=10,
                    alpha_matting_base_size=1000)
        f.close()
        f = open(out_img_path, "wb")
        f.write(img)
        f.close()
    """



from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from yta_multimedia.video.edition.effect.utils.resize_t_functions import linear_zoom_out_t_func
from moviepy.editor import CompositeVideoClip, VideoFileClip, ImageClip, VideoClip, ColorClip
from typing import Union


class LinearZoomOutVideoEffect(VideoEffect):
    """
    Creates a linear Zoom out effect in the provided video.
    """
    
    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], zoom_ratio = None):
        """
        Applies the effect on the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)
        # We reset the mask to avoid problems with zoom
        video = video.add_mask()

        if zoom_ratio is None:
            zoom_ratio = 0.2

        fps = video.fps
        duration = video.duration
        screensize = video.size

        effected_video = (
            video
            .resize(screensize)
            .resize(lambda t: linear_zoom_out_t_func(t, duration, zoom_ratio))
            .set_position(('center', 'center'))
            .set_duration(duration)
            .set_fps(fps)
        )

        return CompositeVideoClip([effected_video], size = screensize)


    


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, AudioFileClip, ColorClip, vfx, VideoClip
from typing import Union
from yta_multimedia.resources.video.effect.sound.drive_urls import SAD_MOMENT_GOOGLE_DRIVE_DOWNLOAD_URL
from yta_multimedia.resources import get_resource
from yta_multimedia.video.edition.effect.constants import EFFECTS_RESOURCES_FOLDER


class SadMomentVideoEffect(VideoEffect):
    """
    This method gets the first frame of the provided 'clip' and returns a
    new clip that is an incredible 'sad_moment' effect with black and white
    filter, zoom in and rotating effect and also sad violin music.

    The 'duration' parameter is to set the returned clip duration, but the
    default value is a perfect one.
    """
    
    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], duration = None):
        """
        Applies the effect to the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        if not duration:
            duration = 4.8

        # We freeze the first frame
        aux = ImageClip(video.get_frame(0), duration = duration)
        aux.fps = video.fps
        video = aux

        # We then build the whole effect
        video = video.fx(vfx.blackwhite).resize(lambda t: 1 + 0.30 * (t / video.duration)).set_position(lambda t: (-(0.15 * video.w * (t / video.duration)), -(0.15 * video.h * (t / video.duration)))).rotate(lambda t: 5 * (t / video.duration), expand = False)

        # We set the effect audio
        TMP_FILENAME = get_resource(SAD_MOMENT_GOOGLE_DRIVE_DOWNLOAD_URL, EFFECTS_RESOURCES_FOLDER + 'sounds/sad_moment.mp3')
        video.audio = AudioFileClip(TMP_FILENAME).set_duration(video.duration)

        return CompositeVideoClip([
            ColorClip(color = [0, 0, 0], size = video.size, duration = video.duration),
            video,
        ])


from yta_multimedia.video.generation.manim.constants import MANDATORY_CONFIG_PARAMETER
from yta_multimedia.video.generation.manim.utils.config import read_manim_config_file, write_manim_config_file
from yta_general_utils.file.filename import get_file_extension, get_file_filename_without_extension
from yta_general_utils.file import rename_file
from yta_general_utils.programming.path import get_code_abspath, get_project_abspath, get_code_filename
from manim.cli.render.commands import render as manim_render
from manim import Scene, config
from threading import Thread


class BaseManimAnimation(Scene):
    """
    General class so that our own classes can inherit it 
    and work correctly.
    """
    required_parameters = {}
    parameters = {}

    def parameter_is_mandatory(self, parameter, required_parameters):
        """
        Returns true if the provided 'parameter' is mandatory, based on
        'required_parameters' definition.
        """
        if parameter in required_parameters and required_parameters[parameter] == MANDATORY_CONFIG_PARAMETER:
            return True
        
        return False

    def __set_mandatory_config(self):
        """
        This method set some configuration parameters we need to build perfect
        animation videos.
        """
        # Disables caching to avoid error when cache is overload
        config.disable_caching = True
        config.max_files_cached = 9999
        # This makes the video background transparent to fit well over the main video
        self.camera.background_opacity = 0.0
    
    def setup(self):
        """
        This method is called when manim is trying to use it to
        render the scene animation. It is called the first, to
        instantiate it and before the 'construct' method that
        is the one that will render.
        """
        self.__set_mandatory_config()
        self.parameters = read_manim_config_file()

        return self.parameters
    
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.setup()

    def animate(self):
        pass

    def generate(self, parameters, renderer: str = 'cairo', output_filename: str = 'base_manim_animation.mov'):
        """
        Generates the animation video file using the provided
        'parameters' and stores it locally as 'output_filename'
        """
        if not output_filename:
            raise Exception('No "output_filename" provided.')
        
        # We write parameters in file to be able to read them
        write_manim_config_file(parameters)

        # Variables we need to make it work
        FPS = str(60)
        CLASS_MANIM_CONTAINER_ABSPATH = get_code_abspath(self.__class__)
        CLASS_FILENAME_WITHOUT_EXTENSION = get_file_filename_without_extension(get_code_filename(self.__class__))
        CLASS_NAME = self.__class__.__name__
        
        output_filename_extension = get_file_extension(output_filename)

        if not renderer or renderer not in ['cairo', 'opengl']:
            renderer = 'cairo'

        # These args are in 'manim.cli.render.commands.py' injected
        # as '@output_options', '@render_options', etc.
        args = {
            # I never used this 'format' before
            '--format': True,
            output_filename_extension.replace('.', ''): True, # Valid values are: [png|gif|mp4|webm|mov]
            # Qualities are here: manim\constants.py > QUALITIES
            '--quality': True,
            'h': True,
            '--fps': True,
            FPS: True,
            '--transparent': True,
            '--renderer': True,
            # The 'cairo' default option has been working good always
            renderer: True, # 'opengl' or 'cairo', 'cairo' is default
            # The '--output_file' changes the created file name, not the path
            CLASS_MANIM_CONTAINER_ABSPATH: True,
            CLASS_NAME: True
        }

        # TODO: Do more Exception checkings (such as '.smtg')
        if output_filename_extension != '.mov':
            del args['--transparent']

        # We need to execute this as a thread because the program ends when
        # finished if not a thread
        render_thread = Thread(target = manim_render, args = [args])
        render_thread.start()
        render_thread.join()
            
        CREATED_FILE_ABSPATH = get_project_abspath() + 'media/videos/' + CLASS_FILENAME_WITHOUT_EXTENSION + '/1080p' + FPS + '/' + CLASS_NAME + output_filename_extension

        rename_file(CREATED_FILE_ABSPATH, output_filename)

        return output_filename


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, VideoClip, ColorClip
from typing import Union
from skimage.filters import gaussian


class BlurVideoEffect(VideoEffect):
    """
    This effect will zoom out the clip, on the center.

    TODO: This effect is not smooth as it makes it have
    a black border. Maybe removing it (?)
    """

    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], blur_radius = None):
        """
        Applies the effect to the provided 'video'.
        """
        video = VideoEffect.parse_moviepy_video(video)

        if blur_radius is None:
            blur_radius = 4

        return video.fl(lambda get_frame, t: gaussian(get_frame(t).astype(float), sigma = blur_radius))



from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_center import CoordinateCenter
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_corner import CoordinateCorner
from yta_multimedia.video.edition.effect.moviepy.position.enums import ScreenPosition
from yta_multimedia.video.edition.effect.moviepy.position.utils.position import position_video_in
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, VideoClip
from typing import Union


class StayAtPositionMoviepyEffect(BasePositionMoviepyEffect):
    """
    Effect of keeping the given 'video' inmobile in a specific
    position given (if given as parameter when applying) or randomly
    generated (inside the bounds according to the also provided
    'background_video' dimensions).
    """

    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], position: Union[ScreenPosition, CoordinateCorner, CoordinateCenter] = ScreenPosition.RANDOM_INSIDE):
        return cls.apply_over_video(video, BasePositionMoviepyEffect.get_black_background_clip(video.duration), position)

    @classmethod
    def apply_over_video(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], background_video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], position: Union[ScreenPosition, CoordinateCorner, CoordinateCenter] = ScreenPosition.RANDOM_INSIDE):
        BasePositionMoviepyEffect.validate_position(position)

        background_video = BasePositionMoviepyEffect.prepare_background_clip(background_video, video)

        effect = position_video_in(video, background_video, position).set_start(0).set_duration(video.duration)

        return CompositeVideoClip([
            background_video,
            effect
        ])


from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_center import CoordinateCenter
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate_corner import CoordinateCorner
from yta_multimedia.video.edition.effect.moviepy.position.enums import ScreenPosition
from yta_multimedia.video.edition.effect.moviepy.position.utils.position import get_moviepy_position
from yta_multimedia.video.edition.effect.moviepy.position.utils.move import circular_movement
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, VideoClip, ColorClip
from typing import Union


class CirclesAtPositionMoviepyEffect(BasePositionMoviepyEffect):
    """
    Effect of moving in circles surrounding the specified position.

    TODO: Implement the radius parameter option
    """

    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], position: Union[ScreenPosition, CoordinateCorner, CoordinateCenter] = ScreenPosition.RANDOM_INSIDE):
        return cls.apply_over_video(video, BasePositionMoviepyEffect.get_black_background_clip(video.duration), position)

    @classmethod
    def apply_over_video(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], background_video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], position: Union[ScreenPosition, CoordinateCorner, CoordinateCenter] = ScreenPosition.RANDOM_INSIDE):
        BasePositionMoviepyEffect.validate_position(position)

        background_video = BasePositionMoviepyEffect.prepare_background_clip(background_video, video)

        position = get_moviepy_position(video, background_video, position)

        effect = video.set_position(lambda t: circular_movement(t, position[0], position[1])).set_start(0).set_duration(video.duration)

        return CompositeVideoClip([
            background_video,
            effect
        ])